import json
import os
import socket
import unittest
from contextlib import ExitStack
from typing import Optional, List, Callable

import requests

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient

INTEGRATION_TEST_DOCKER_ENVIRONMENT_DEFAULT_BIN = "./start-test-env"


def find_free_ports(num_ports: int) -> List[int]:

    ret_val = list()
    with ExitStack() as stack:
        sockets = [stack.enter_context(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) for dummy in range(num_ports)]
        for s in sockets:
            s.bind(('', 0))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            ret_val.append(s.getsockname()[1])
    with ExitStack() as stack:
        # Create an array of tuples of new socket + port to use
        sockets = [(stack.enter_context(socket.socket(socket.AF_INET, socket.SOCK_STREAM)), port) for port in ret_val]
        for socket_port in sockets:
            s = socket_port[0]
            port = socket_port[1]
            s.bind(('', port))
            s.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
    return ret_val


def close_environments(*args):
    for env in args:
        try:
            if env is not None:
                env.close()
        except Exception as e:
            print(e)


def check_db_version_from_env() -> Optional[str]:
    retval = None
    if "EXASOL_VERSION" in os.environ and os.environ["EXASOL_VERSION"] != "default":
        retval = os.environ["EXASOL_VERSION"]
    return retval


def multiassert(assert_list: List[Callable], unit_test: unittest.TestCase):
    failure_log: List[str] = []
    for assert_fn in assert_list:
        try:
            assert_fn()
        except AssertionError as e:
            failure_log.append(f"\nFailure {len(failure_log)}: {str(e)}")

    if len(failure_log) != 0:
        res_failure_log = '\n'.join(failure_log)
        unit_test.fail(f"{len(failure_log)} failures within test.\n {res_failure_log}")



import inspect
import os
import shutil
import tempfile
from pathlib import Path
from sys import stderr
from typing import Dict, Any

from exasol_integration_test_docker_environment.lib.api import spawn_test_environment
from exasol_integration_test_docker_environment.lib.api import spawn_test_environment_with_test_container
from exasol_integration_test_docker_environment.lib.data.test_container_content_description import \
    TestContainerContentDescription
from exasol_integration_test_docker_environment.testing.docker_registry import default_docker_repository_name
from exasol_integration_test_docker_environment \
    .testing.exaslct_docker_test_environment import \
    ExaslctDockerTestEnvironment
from exasol_integration_test_docker_environment \
    .testing.exaslct_test_environment import (
        get_class,
        get_test_flavor,
)
from exasol_integration_test_docker_environment.testing.utils import find_free_ports


class ApiTestEnvironment:

    def __init__(self, test_object, name=None):
        self.test_object = test_object
        self.test_class = get_class(test_object)
        self.flavor_path = get_test_flavor(self.test_class)
        self.name = name if name else self.test_class.__name__
        self.docker_repository_name = default_docker_repository_name(self.name)
        self.temp_dir = tempfile.mkdtemp()

    @property
    def output_dir(self):
        return self.temp_dir

    @property
    def task_dependency_dot_file(self):
        return f"{self.name}.dot"

    def close(self):
        try:
            shutil.rmtree(self.temp_dir)
        except Exception as e:
            print(e, file=stderr)

    def _get_default_test_environment(self, name: str, database_port: int, bucketfs_port: int):
        return ExaslctDockerTestEnvironment(
            name=self.name + "_" + name,
            database_host="localhost",
            db_username="sys",
            db_password="exasol",
            bucketfs_username="w",
            bucketfs_password="write",
            database_port=database_port,
            bucketfs_port=bucketfs_port)

    def spawn_docker_test_environment_with_test_container(self, name: str,
                                                          test_container_content: TestContainerContentDescription,
                                                          additional_parameter: Dict[str, Any] = None) \
            -> ExaslctDockerTestEnvironment:
        if additional_parameter is None:
            additional_parameter = dict()
        database_port, bucketfs_port = find_free_ports(2)
        on_host_parameter = self._get_default_test_environment(name, database_port, bucketfs_port)
        docker_db_image_version = on_host_parameter.docker_db_image_version
        on_host_parameter.environment_info, on_host_parameter.clean_up = \
            spawn_test_environment_with_test_container(environment_name=on_host_parameter.name,
                                                       database_port_forward=on_host_parameter.database_port,
                                                       bucketfs_port_forward=on_host_parameter.bucketfs_port,
                                                       docker_db_image_version=docker_db_image_version,
                                                       test_container_content=test_container_content,
                                                       **additional_parameter)
        return on_host_parameter

    def spawn_docker_test_environment(self, name: str,
                                      additional_parameter: Dict[str, Any] = None) \
            -> ExaslctDockerTestEnvironment:
        if additional_parameter is None:
            additional_parameter = dict()
        database_port, bucketfs_port = find_free_ports(2)
        on_host_parameter = self._get_default_test_environment(name, database_port, bucketfs_port)
        on_host_parameter.environment_info, on_host_parameter.clean_up = \
            spawn_test_environment(environment_name=on_host_parameter.name,
                                   database_port_forward=on_host_parameter.database_port,
                                   bucketfs_port_forward=on_host_parameter.bucketfs_port,
                                   docker_db_image_version=on_host_parameter.docker_db_image_version,
                                   **additional_parameter)
        return on_host_parameter



import click
from exasol_integration_test_docker_environment \
    .lib.test_environment.parameter \
    .docker_db_test_environment_parameter \
    import DbOsAccess

test_environment_options = [
    click.option('--environment-type', type=click.Choice(['docker_db', 'external_db']), default="""docker_db""",
                 show_default=True,
                 help="""Environment type for tests."""),
    click.option('--max_start_attempts', type=int, default=2,
                 show_default=True,
                 help="""Maximum start attempts for environment""")

]

LATEST_DB_VERSION = """8.18.1"""

docker_db_options = [
    click.option('--docker-db-image-version', type=str, default=LATEST_DB_VERSION,
                 show_default=True,
                 help="""Docker DB Image Version against which the tests should run."""),
    click.option('--docker-db-image-name', type=str, default="""exasol/docker-db""",
                 show_default=True,
                 help="""Docker DB Image Name against which the tests should run."""),
    click.option('--db-os-access', type=click.Choice([e.name for e in DbOsAccess]),
                 metavar="METHOD", default="""DOCKER_EXEC""", show_default=True,
                 help="""How to access file system and command line of the
     		 database operating system. Experimental option, will show no
     		 effect until implementation of feature SSH access is
     		 completed."""),
    click.option('--create-certificates/--no-create-certificates', default=False,
                 help="""Creates and injects SSL certificates to the Docker DB container."""),
    click.option('--additional-db-parameter', '-p', type=str, multiple=True,
                 help="""Additional database parameter which will be injected to EXAConf. Value should have format '-param=value'.""")
]

external_db_options = [
    click.option('--external-exasol-db-host', type=str,
                 help="""Host name or IP of external Exasol DB, needs to be set if --environment-type=external_db"""),
    click.option('--external-exasol-db-port', type=int, default=8563,
                 help="""Database port of external Exasol DB, needs to be set if --environment-type=external_db"""),
    click.option('--external-exasol-bucketfs-port', type=int, default=6583,
                 help="""Bucketfs port of external Exasol DB, needs to be set if --environment-type=external_db"""),
    click.option('--external-exasol-db-user', type=str,
                 help="""User for external Exasol DB, needs to be set if --environment-type=external_db"""),
    click.option('--external-exasol-db-password', type=str,
                 help="""Database Password for external Exasol DB"""),
    click.option('--external-exasol-bucketfs-write-password', type=str,
                 help="""BucketFS write Password for external Exasol DB"""),
    click.option('--external-exasol-xmlrpc-host', type=str,
                 help="""Hostname for the xmlrpc server"""),
    click.option('--external-exasol-xmlrpc-port', type=int,
                 default=443, show_default=True,
                 help="""Port for the xmlrpc server"""),
    click.option('--external-exasol-xmlrpc-user', type=str,
                 default="""admin""", show_default=True,
                 help="""User for the xmlrpc server"""),
    click.option('--external-exasol-xmlrpc-password', type=str,
                 help="""Password for the xmlrpc server"""),
    click.option('--external-exasol-xmlrpc-cluster-name', type=str,
                 default="""cluster1""", show_default=True,
                 help="""Password for the xmlrpc server""")

]



import base64
import hashlib
from pathlib import Path, PurePath
from typing import Dict

# TODO add hash config to the hash
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.file_directory_list_hasher import \
    FileDirectoryListHasher, PathMapping
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageDescription, ImageInfo


class BuildContextHasher:

    def __init__(self, logger, image_description: ImageDescription):
        self.image_description = image_description
        self.logger = logger

    def generate_image_hash(self, image_info_of_dependencies: Dict[str, ImageInfo]):
        hash_of_build_context = self._generate_build_context_hash()
        final_hash = self._generate_final_hash(hash_of_build_context, image_info_of_dependencies)
        return self._encode_hash(final_hash)

    def _generate_build_context_hash(self):
        files_directories_list_hasher = \
            FileDirectoryListHasher(followlinks=True,
                                    hashfunc="sha256",
                                    hash_file_names=True,
                                    hash_directory_names=True,
                                    hash_permissions=True)
        files_directories_to_hash = [PathMapping(PurePath(destination), Path(source)) for destination, source in
                                     self.image_description.mapping_of_build_files_and_directories.items()]
        # Use only the dockerfile itself for hashing. In order to accomplish that,
        # set the dockerfile only as destination in the mapping
        dockerfile = Path(self.image_description.dockerfile).name
        files_directories_to_hash.append(PathMapping(PurePath(dockerfile),
                                                     Path(self.image_description.dockerfile)))
        self.logger.debug("files_directories_list_hasher %s", files_directories_to_hash)
        hash_of_build_context = files_directories_list_hasher.hash(files_directories_to_hash)
        self.logger.debug("hash_of_build_context %s", self._encode_hash(hash_of_build_context))
        return hash_of_build_context

    def _generate_final_hash(self, hash_of_build_context: bytes,
                             image_info_of_dependencies: Dict[str, ImageInfo]):
        hasher = hashlib.sha256()
        self.add_image_changing_build_arguments(hasher)
        self.add_dependencies(hasher, image_info_of_dependencies)
        hasher.update(hash_of_build_context)
        final_hash = hasher.digest()
        return final_hash

    def add_dependencies(self, hasher, image_info_of_dependencies: Dict[str, ImageInfo]):
        hashes_of_dependencies = \
            [(key, image_info.hash) for key, image_info in image_info_of_dependencies.items()]
        hashes_to_hash = sorted(hashes_of_dependencies, key=lambda t: t[0])
        self.logger.debug("hashes_to_hash %s", hashes_to_hash)
        for key, image_name in hashes_to_hash:
            hasher.update(key.encode("utf-8"))
            hasher.update(image_name.encode("utf-8"))

    def add_image_changing_build_arguments(self, hasher):
        arguments = \
            [(key, value) for key, value
             in self.image_description.image_changing_build_arguments.items()]
        sorted_arguemnts = sorted(arguments, key=lambda t: t[0])
        for key, value in sorted_arguemnts:
            hasher.update(key.encode("utf-8"))
            hasher.update(value.encode("utf-8"))

    def _encode_hash(self, hash_of_build_directory):
        base32 = base64.b32encode(hash_of_build_directory)
        ascii = base32.decode("ASCII")
        trim = ascii.replace("=", "")
        return trim



import click

DEFAULT_OUTPUT_DIRECTORY = ".build_output"

output_directory_option = click.option('--output-directory', type=click.Path(file_okay=False, dir_okay=True),
                                       default=DEFAULT_OUTPUT_DIRECTORY,
                                       show_default=True,
                                       help="Output directory where the system stores all output and log files.")

tempory_base_directory_option = click.option('--temporary-base-directory',
                                             type=click.Path(file_okay=False,
                                                             dir_okay=True),
                                             default="/tmp",
                                             show_default=True,
                                             help="Directory where the system creates temporary directories.")

system_options = [
    click.option('--workers', type=int,
                 default=5, show_default=True,
                 help="Number of parallel workers"),
    click.option('--task-dependencies-dot-file', type=click.Path(file_okay=True),
                 default=None, help="Path where to store the Task Dependency Graph as dot file"),
]

luigi_logging_options = [
    click.option('--log-level', type=click.Choice(['DEBUG', 'INFO', 'WARNING', 'ERROR', 'FATAL']),
                 default=None, help="Log level used for console logging"),
    click.option('--use-job-specific-log-file', type=bool,
                 default=True, help="Use a job specific log file which write the debug log "
                                    "to the job directory in the build directory")
]



import json

from exasol_integration_test_docker_environment.lib.config.log_config import WriteLogFilesToConsole
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo
from exasol_integration_test_docker_environment.lib.logging.abstract_log_handler import AbstractLogHandler


class PullLogHandler(AbstractLogHandler):

    def __init__(self, log_file_path, logger, image_info: ImageInfo):
        super().__init__(log_file_path, logger)
        self._image_info = image_info

    def handle_log_line(self, log_line, error: bool = False):
        json_output = json.loads(log_line)
        if "status" in json_output \
                and json_output["status"] != "Downloading" \
                and json_output["status"] != "Extracting":
            self._complete_log.append(json_output["status"])
            self._log_file.write(json_output["status"])
            self._log_file.write("\n")
            self._log_file.flush()
        if 'errorDetail' in json_output:
            self._error_message = json_output["errorDetail"]["message"]

    def finish(self):
        self.write_log_to_console_if_requested()
        if self._error_message is not None:
            self.write_error_log_to_console_if_requested()
            raise Exception(
                "Error occurred during the pull of the image %s. Received error \"%s\" ."
                "The whole log can be found in %s"
                % (self._image_info.get_source_complete_name(), self._error_message, self._log_file_path))

    def write_error_log_to_console_if_requested(self):
        if self._log_config.write_log_files_to_console == WriteLogFilesToConsole.only_error:
            self._logger.error("Pull of image %s failed\nPush Log:\n%s",
                               self._image_info.get_source_complete_name(),
                               "\n".join(self._complete_log))

    def write_log_to_console_if_requested(self):
        if self._log_config.write_log_files_to_console == WriteLogFilesToConsole.all:
            self._logger.info("Pull Log of image %s\n%s",
                              self._image_info.get_source_complete_name(),
                              "\n".join(self._complete_log))



import shutil
import tempfile
from pathlib import Path

from exasol_integration_test_docker_environment.lib.base.still_running_logger import StillRunningLogger
from exasol_integration_test_docker_environment.lib.config.build_config import build_config
from exasol_integration_test_docker_environment.lib.config.docker_config import docker_build_arguments
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_creator_base_task import \
    DockerImageCreatorBaseTask
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.build_context_creator import \
    BuildContextCreator
from exasol_integration_test_docker_environment.lib.docker.images.create.utils.build_log_handler import BuildLogHandler
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo


class DockerBuildImageTask(DockerImageCreatorBaseTask):

    def run_task(self):
        self.logger.info("Build docker image %s, log file can be found here %s",
                         self.image_info.get_target_complete_name(), self.get_log_path())
        temp_directory = tempfile.mkdtemp(prefix="script_language_container_tmp_dir",
                                          dir=build_config().temporary_base_directory)
        try:
            image_description = self.image_info.image_description
            build_context_creator = BuildContextCreator(temp_directory,
                                                        self.image_info,
                                                        self.get_log_path())
            build_context_creator.prepare_build_context_to_temp_dir()
            with self._get_docker_client() as docker_client:
                output_generator = \
                    docker_client.api.build(path=temp_directory,
                                            nocache=self.no_cache,
                                            tag=self.image_info.get_target_complete_name(),
                                            rm=True,
                                            buildargs=dict(**image_description.transparent_build_arguments,
                                                           **image_description.image_changing_build_arguments,
                                                           **docker_build_arguments().secret))
                self._handle_output(output_generator, self.image_info)
        finally:
            shutil.rmtree(temp_directory)

    def _handle_output(self, output_generator, image_info: ImageInfo):
        log_file_path = Path(self.get_log_path(), "docker-build.log")
        with BuildLogHandler(log_file_path, self.logger, image_info) as log_handler:
            still_running_logger = StillRunningLogger(
                self.logger, "build image %s" % image_info.get_target_complete_name())
            for log_line in output_generator:
                still_running_logger.log()
                log_handler.handle_log_lines(log_line)



import functools
from typing import Tuple, Optional, Callable
import humanfriendly

from exasol_integration_test_docker_environment.lib.api.common import set_build_config, set_docker_repository_config, \
    run_task, generate_root_task, no_cli_function
from exasol_integration_test_docker_environment.cli.options.docker_repository_options import \
    DEFAULT_DOCKER_REPOSITORY_NAME
from exasol_integration_test_docker_environment.cli.options.system_options import DEFAULT_OUTPUT_DIRECTORY
from exasol_integration_test_docker_environment.cli.options.test_environment_options import LATEST_DB_VERSION
from exasol_integration_test_docker_environment.lib.api.api_errors import ArgumentConstraintError
from exasol_integration_test_docker_environment.lib.data.environment_info import EnvironmentInfo
from exasol_integration_test_docker_environment.lib.data.test_container_content_description import \
    TestContainerContentDescription
from exasol_integration_test_docker_environment.lib.docker.container.utils import remove_docker_container
from exasol_integration_test_docker_environment.lib.docker.volumes.utils import remove_docker_volumes
from exasol_integration_test_docker_environment.lib.docker.networks.utils import remove_docker_networks
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_environment_with_docker_db import \
    SpawnTestEnvironmentWithDockerDB
from exasol_integration_test_docker_environment \
    .lib.test_environment.parameter \
    .docker_db_test_environment_parameter import DbOsAccess


def _cleanup(environment_info: EnvironmentInfo) -> None:
    if environment_info.test_container_info is None:
        remove_docker_container([environment_info.database_info.container_info.container_name])
    else:
        remove_docker_container([environment_info.test_container_info.container_name,
                                 environment_info.database_info.container_info.container_name])
    remove_docker_volumes([environment_info.database_info.container_info.volume_name])
    remove_docker_networks([environment_info.network_info.network_name])


@no_cli_function
def spawn_test_environment_with_test_container(
        environment_name: str,
        test_container_content: TestContainerContentDescription,
        database_port_forward: Optional[int] = None,
        bucketfs_port_forward: Optional[int] = None,
        db_mem_size: str = "2 GiB",
        db_disk_size: str = "2 GiB",
        nameserver: Tuple[str, ...] = tuple(),
        docker_runtime: Optional[str] = None,
        docker_db_image_version: str = LATEST_DB_VERSION,
        docker_db_image_name: str = "exasol/docker-db",
        db_os_access: str = "DOCKER_EXEC",
        create_certificates: bool = False,
        additional_db_parameter: Tuple[str, ...] = tuple(),
        source_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
        source_docker_tag_prefix: str = '',
        source_docker_username: Optional[str] = None,
        source_docker_password: Optional[str] = None,
        target_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
        target_docker_tag_prefix: str = '',
        target_docker_username: Optional[str] = None,
        target_docker_password: Optional[str] = None,
        output_directory: str = DEFAULT_OUTPUT_DIRECTORY,
        temporary_base_directory: str = "/tmp",
        workers: int = 5,
        task_dependencies_dot_file: Optional[str] = None,
        log_level: Optional[str] = None,
        use_job_specific_log_file: bool = False
) \
        -> Tuple[EnvironmentInfo, Callable[[], None]]:
    """
    This function spawns a test environment with a docker-db container and a connected test-container.
    The test-container is reachable by the database for output redirects of UDFs.
    The function returns an environment_info object, describing the environment, and a cleanup-method, which
    can be used to stop the environment.
    raises: TaskRuntimeError if spawning the test environment fails

    """
    parsed_db_mem_size = humanfriendly.parse_size(db_mem_size)
    if parsed_db_mem_size < humanfriendly.parse_size("1 GiB"):
        raise ArgumentConstraintError("db_mem_size", "needs to be at least 1 GiB")
    parsed_db_disk_size = humanfriendly.parse_size(db_disk_size)
    if parsed_db_disk_size < humanfriendly.parse_size("100 MiB"):
        raise ArgumentConstraintError("db_disk_size", "needs to be at least 100 MiB")
    set_build_config(False,
                     tuple(),
                     False,
                     False,
                     output_directory,
                     temporary_base_directory,
                     None,
                     None)
    set_docker_repository_config(source_docker_password, source_docker_repository_name, source_docker_username,
                                 source_docker_tag_prefix, "source")
    set_docker_repository_config(target_docker_password, target_docker_repository_name, target_docker_username,
                                 target_docker_tag_prefix, "target")
    task_creator = lambda: generate_root_task(task_class=SpawnTestEnvironmentWithDockerDB,
                                              environment_name=environment_name,
                                              database_port_forward=str(
                                                  database_port_forward) if database_port_forward is not None else None,
                                              bucketfs_port_forward=str(
                                                  bucketfs_port_forward) if bucketfs_port_forward is not None else None,
                                              mem_size=db_mem_size,
                                              disk_size=db_disk_size,
                                              nameservers=nameserver,
                                              docker_runtime=docker_runtime,
                                              docker_db_image_version=docker_db_image_version,
                                              docker_db_image_name=docker_db_image_name,
                                              db_os_access=DbOsAccess[db_os_access],
                                              db_user="sys",
                                              db_password="exasol",
                                              bucketfs_write_password="write",
                                              no_test_container_cleanup_after_success=True,
                                              no_test_container_cleanup_after_failure=False,
                                              no_database_cleanup_after_success=True,
                                              no_database_cleanup_after_failure=False,
                                              create_certificates=create_certificates,
                                              test_container_content=test_container_content,
                                              additional_db_parameter=additional_db_parameter
                                              )
    environment_info = run_task(task_creator, workers, task_dependencies_dot_file,
                                log_level=log_level,
                                use_job_specific_log_file=use_job_specific_log_file)
    return environment_info, functools.partial(_cleanup, environment_info)



from typing import List

from exasol_integration_test_docker_environment.lib.base.info import Info
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo


class ContainerInfo(Info):

    def __init__(self, container_name: str,
                 ip_address: str,
                 network_aliases: List[str],
                 network_info: DockerNetworkInfo,
                 volume_name: str = None):
        self.network_aliases = network_aliases
        self.ip_address = ip_address
        self.network_info = network_info
        self.container_name = container_name
        self.volume_name = volume_name



from typing import Optional

from exasol_integration_test_docker_environment.lib.data.container_info import ContainerInfo
from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo
from exasol_integration_test_docker_environment.lib.data.docker_volume_info import DockerVolumeInfo
from exasol_integration_test_docker_environment.lib.data.environment_type import EnvironmentType
from exasol_integration_test_docker_environment.lib.test_environment.abstract_spawn_test_environment import \
    AbstractSpawnTestEnvironment
from exasol_integration_test_docker_environment.lib.test_environment.create_certificates.create_ssl_certificates_task import \
    CreateSSLCertificatesTask
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.wait_for_test_docker_database import \
    WaitForTestDockerDatabase
from exasol_integration_test_docker_environment.lib.test_environment.db_version import \
    db_version_supports_custom_certificates
from exasol_integration_test_docker_environment.lib.test_environment.parameter.docker_db_test_environment_parameter import \
    DockerDBTestEnvironmentParameter
from exasol_integration_test_docker_environment.lib.test_environment.prepare_network_for_test_environment import \
    PrepareDockerNetworkForTestEnvironment
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_database import SpawnTestDockerDatabase


class SpawnTestEnvironmentWithDockerDB(
    AbstractSpawnTestEnvironment,
    DockerDBTestEnvironmentParameter):

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.db_container_name = f"""db_container_{self.environment_name}"""
        self.certificate_volume_name = f"""certificates_{self.environment_name}"""

    def get_environment_type(self):
        return EnvironmentType.docker_db

    def create_ssl_certificates(self):
        if not db_version_supports_custom_certificates(self.docker_db_image_version):
            raise ValueError("Minimal supported Database with custom certificates is '7.0.6'")
        return \
            self.create_child_task_with_common_params(
                CreateSSLCertificatesTask,
                environment_name=self.environment_name,
                db_container_name=self.db_container_name,
                network_name=self.network_name,
                docker_runtime=self.docker_runtime,
                volume_name=self.certificate_volume_name,
                reuse=self.reuse_database or self.reuse_test_container,
                no_cleanup_after_success=self.no_database_cleanup_after_success or self.no_test_container_cleanup_after_success,
                no_cleanup_after_failure=self.no_database_cleanup_after_failure or self.no_test_container_cleanup_after_failure,
            )

    def create_network_task(self, attempt: int):
        return \
            self.create_child_task_with_common_params(
                PrepareDockerNetworkForTestEnvironment,
                test_container_name=self.test_container_name,
                network_name=self.network_name,
                db_container_name=self.db_container_name,
                reuse=self.reuse_database or self.reuse_test_container,
                no_cleanup_after_success=self.no_database_cleanup_after_success or self.no_test_container_cleanup_after_success,
                no_cleanup_after_failure=self.no_database_cleanup_after_failure or self.no_test_container_cleanup_after_failure,
                attempt=attempt
            )

    def create_spawn_database_task(self,
                                   network_info: DockerNetworkInfo,
                                   certificate_volume_info: Optional[DockerVolumeInfo],
                                   attempt: int):
        certificate_volume_name = certificate_volume_info.volume_name if certificate_volume_info is not None else None

        return \
            self.create_child_task_with_common_params(
                SpawnTestDockerDatabase,
                db_container_name=self.db_container_name,
                network_info=network_info,
                certificate_volume_name=certificate_volume_name,
                ip_address_index_in_subnet=0,
                attempt=attempt,
                additional_db_parameter=self.additional_db_parameter
            )

    def create_wait_for_database_task(self,
                                      attempt: int,
                                      database_info: DatabaseInfo):
        return \
            self.create_child_task_with_common_params(
                WaitForTestDockerDatabase,
                database_info=database_info,
                attempt=attempt,
                docker_db_image_version=self.docker_db_image_version)



import time

from exasol_integration_test_docker_environment.lib.test_environment.database_setup.bucketfs_sync_checker import \
    BucketFSSyncChecker


class TimeBasedBucketFSSyncWaiter(BucketFSSyncChecker):

    def __init__(self, sync_time_estimation):
        self.sync_time_estimation = sync_time_estimation

    def prepare_upload(self):
        pass

    def wait_for_bucketfs_sync(self):
        time.sleep(self.sync_time_estimation)



import time
from pathlib import PurePath
from threading import Thread

from docker.models.containers import Container

from exasol_integration_test_docker_environment.lib.data.database_credentials import DatabaseCredentials
from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo
from exasol_integration_test_docker_environment.lib.test_environment.database_setup.find_exaplus_in_db_container import \
    find_exaplus


class IsDatabaseReadyThread(Thread):

    def __init__(self,
                 logger,
                 database_info: DatabaseInfo,
                 database_container: Container,
                 database_credentials: DatabaseCredentials,
                 docker_db_image_version: str):
        super().__init__()
        self.logger = logger
        self.database_credentials = database_credentials
        self._database_info = database_info
        self._db_container = database_container
        self.finish = False
        self.is_ready = False
        self.output_db_connection = None
        self.output_bucketfs_connection = None
        self.docker_db_image_version = docker_db_image_version

    def stop(self):
        self.logger.info("Stop IsDatabaseReadyThread")
        self.finish = True

    def run(self):
        db_connection_command = ""
        bucket_fs_connection_command = ""
        try:
            exaplus_path = find_exaplus(self._db_container)
            db_connection_command = self.create_db_connection_command(exaplus_path)
            bucket_fs_connection_command = self.create_bucketfs_connection_command()
        except RuntimeError as e:
            self.logger.error(e)
            self.finish = True
        while not self.finish:
            (exit_code_db_connection, self.output_db_connection) = \
                self._db_container.exec_run(cmd=db_connection_command)
            (exit_code_bucketfs_connection, self.output_bucketfs_connection) = \
                self._db_container.exec_run(cmd=bucket_fs_connection_command)
            if exit_code_db_connection == 0 and exit_code_bucketfs_connection == 0:
                self.finish = True
                self.is_ready = True
            time.sleep(1)

    def create_db_connection_command(self, exaplus_path: PurePath):
        username = self.database_credentials.db_user
        password = self.database_credentials.db_password
        connection_options = f"""-c 'localhost:{self._database_info.db_port}' -u '{username}' -p '{password}'"""

        cmd = f"""{exaplus_path} {connection_options}  -sql 'select 1;' -jdbcparam 'validateservercertificate=0'"""
        bash_cmd = f"""bash -c "{cmd}" """
        return bash_cmd

    def create_bucketfs_connection_command(self):
        username = "w"
        password = self.database_credentials.bucketfs_write_password
        cmd = f"""curl --silent --show-error --fail '{username}:{password}@localhost:{self._database_info.bucketfs_port}'"""
        bash_cmd = f"""bash -c "{cmd}" """
        return bash_cmd



import json

import jsonpickle

from exasol_integration_test_docker_environment.lib.base.frozendict_to_dict import FrozenDictToDict


class Info:
    def to_json(self, indent=4):
        jsonpickle.set_preferred_backend('simplejson')
        jsonpickle.set_encoder_options('simplejson', sort_keys=True, indent=indent)
        return jsonpickle.encode(self)

    def to_dict(self):
        return json.loads(self.to_json())

    def __repr__(self):
        return f"{self.__class__.__name__}({self.__dict__})"

    @classmethod
    def from_json(cls, json_string):
        loaded_object = jsonpickle.decode(json_string)
        if not isinstance(loaded_object, cls):
            raise TypeError("Type %s of loaded object does not match %s" % (type(loaded_object), cls))
        return loaded_object

    @classmethod
    def from_dict(cls, dictionary):
        converted_dictionary = FrozenDictToDict().convert(dictionary)
        dumps = json.dumps(converted_dictionary)
        return cls.from_json(dumps)



import sys
from datetime import datetime
from traceback import print_tb

from exasol_integration_test_docker_environment.lib.api.api_errors import TaskRuntimeError


def print_err(*args, **kwargs):
    kwargs["file"] = sys.stderr
    print(*args, **kwargs)


class TerminationHandler:
    """
    This helper class measures and logs time duration of the job and also logs the error message.
    """

    def __init__(self):
        self._start_time = None

    def __enter__(self):
        self._start_time = datetime.now()

    def __exit__(self, exc_type, exc_val, exc_tb):
        exit_value = 0
        if exc_type == TaskRuntimeError:
            self._handle_failure(exc_val)
            exit_value = 1
        elif exc_type is None and exc_val is None and exc_tb is None:
            self._handle_success()
        else:
            self._handle_unexpected_failure(exc_val, exc_tb)
            exit_value = 1
        sys.exit(exit_value)

    def _handle_unexpected_failure(self, exc_val, exc_tb):
        timedelta = datetime.now() - self._start_time
        print_err("The command failed after %s s with:" % timedelta.total_seconds())
        print_err("Caught exception:%s" % exc_val)
        print_tb(exc_tb, file=sys.stderr)

    def _handle_failure(self, task_error: TaskRuntimeError):
        timedelta = datetime.now() - self._start_time
        print_err("The command failed after %s s with:" % timedelta.total_seconds())
        self._print_task_failures(task_error)

    @staticmethod
    def _print_task_failures(task_error: TaskRuntimeError):
        print_err()
        print_err("Task failure message: %s" % task_error.msg)
        print_err(task_error.__cause__.args[0])
        print_err()

    def _handle_success(self):
        timedelta = datetime.now() - self._start_time
        print_err("The command took %s s" % timedelta.total_seconds())



import luigi
from luigi import Config

from exasol_integration_test_docker_environment.lib.test_environment.parameter.test_container_parameter import \
    OptionalTestContainerParameter


class GeneralSpawnTestEnvironmentParameter(OptionalTestContainerParameter):
    reuse_database_setup = luigi.BoolParameter(False, significant=False)
    reuse_test_container = luigi.BoolParameter(False, significant=False)
    no_test_container_cleanup_after_success = luigi.BoolParameter(False, significant=False)
    no_test_container_cleanup_after_failure = luigi.BoolParameter(False, significant=False)
    max_start_attempts = luigi.IntParameter(2, significant=False)
    docker_runtime = luigi.OptionalParameter(None, significant=False)
    create_certificates = luigi.BoolParameter()
    additional_db_parameter = luigi.ListParameter()



import logging
from typing import List

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient


def remove_docker_volumes(volumes: List[str]):
    """
    Removes the given volumes using docker API.
    """
    with ContextDockerClient() as docker_client:
        for volume in volumes:
            try:
                docker_client.volumes.get(volume).remove(force=True)
            except Exception as e:
                logging.error(e)



import click

push_options = [
    click.option('--force-push/--no-force-push', default=False,
                 help="Forces the system to overwrite existing images in registry for build steps that run"),
    click.option('--push-all/--no-push-all', default=False,
                 help="Forces the system to push all images of build-steps that are specified by the goals")

]



from exasol_integration_test_docker_environment.lib.api.health import health
from exasol_integration_test_docker_environment.lib.api.build_test_container import build_test_container
from exasol_integration_test_docker_environment.lib.api.push_test_container import push_test_container
from exasol_integration_test_docker_environment.lib.api.spawn_test_environment import spawn_test_environment
from exasol_integration_test_docker_environment.lib.api.spawn_test_environment_with_test_container import spawn_test_environment_with_test_container



import click

from exasol_integration_test_docker_environment.cli.options.system_options import tempory_base_directory_option, \
    output_directory_option

build_options = [
    click.option('--force-rebuild/--no-force-rebuild', default=False,
                 help="Forces the system to complete rebuild all stages down to the stages "
                      "specified with the options --force-rebuild-from."),
    click.option('--force-rebuild-from', multiple=True, type=str,
                 help="If the option --force-rebuild is given, "
                      "this options specifies for which stages and dependent stages system will force a rebuild. "
                      "The option can be repeated with different stages. "
                      "The system will than force the rebuild of these stages and their. dependet stages."
                 ),
    click.option('--force-pull/--no-force-pull', default=False,
                 help="Forces the system to pull all stages if available, otherwise it rebuilds a stage."),
    output_directory_option,
    tempory_base_directory_option,
    click.option('--log-build-context-content/--no-log-build-context-content',
                 default=False,
                 help="For Debugging: Logs the files and directories in the build context of a stage"),
    click.option('--cache-directory', default=None, type=click.Path(file_okay=False, dir_okay=True, exists=False),
                 help="Directory from where saved docker images can be loaded"),
    click.option('--build-name', default=None, type=str,
                 help="Name of the build. For example: Repository + CI Build Number"),
]



from pathlib import PurePath

import docker.models.containers


def find_exaplus(db_container: docker.models.containers.Container) -> PurePath:
    """
    Tries to find path of exaplus in given container in directories where exaplus is normally installed.
    :db_container Container where to search for exaplus
    """
    exit, output = db_container.exec_run(cmd="find /usr/opt -name 'exaplus' -type f")
    if exit != 0 or output == b"":
        # Using /usr/opt and /opt together in one command doesn't work, because in Exasol 8, /usr/opt doesn't exist
        # and as such the command fails, even if it finds exaplus in /opt
        exit, output = db_container.exec_run(cmd="find /opt -name 'exaplus' -type f")
    if exit != 0:
        raise RuntimeError(f"Exaplus not found on docker db! Output is {output}")
    found_paths = list(filter(None, output.decode("UTF-8").split("\n")))
    if len(found_paths) != 1:
        raise RuntimeError(f"Error determining exaplus path! Output is {output}")
    exaplus_path = PurePath(found_paths[0])
    exit, output = db_container.exec_run(cmd=f"{exaplus_path} --help")
    if exit != 0:
        raise RuntimeError(f"Exaplus not working as expected! Output is {output}")
    return exaplus_path



import inspect
from typing import Any, List, Tuple

import click


def adjust_default_value_for_multiple(x: Any):
    """
    Click stores default values as list if 'Multiple'=true. However, for plain Python methods we
    need to use (immutable) tuples for declaring default value. Hence, we need to convert lists with tuples
    for comparison.
    Also, default values for none-required parameters of with multiple=true are marked as "None"
    in the command object, but in reality click invokes the command with "tuple()" for that parameter.
    (Note that we use here internal structures of click and don't have guarantee of the behavior)
    """
    def_value = x.default
    if x.multiple:
        if type(def_value) == list:
            return tuple(def_value)
        elif def_value is None:
            return tuple()
    return def_value


def is_click_command(obj: Any) -> bool:
    """
    Returns True if parameter obj is a click command, False otherwise.
    """
    return isinstance(obj, click.Command)


def defaults_of_click_call(click_call: click.Command) -> List[Tuple[str, Any]]:
    """
    Returns the default values of all None-required parameters of a click-command.
    """
    return [(o.name, adjust_default_value_for_multiple(o)) for o in click_call.params if not o.required]


def param_names_of_click_call(click_call: click.Command) -> List[str]:
    """
    Returns names of all parameters of a click call
    """
    return [o.name for o in click_call.params]


def get_click_and_api_functions(click_module, api_module) -> Tuple[List[Any], List[Any]]:
    # Get all click commands in module 'click_module'
    click_commands = [c[1] for c in inspect.getmembers(click_module, is_click_command)]
    # Get all functions in module 'api_module'
    api_functions = [f[1] for f in inspect.getmembers(api_module, inspect.isfunction)
                     if f[1].__cli_function__]
    return click_commands, api_functions


def get_click_and_api_function_names(click_module, api_module) -> Tuple[List[Any], List[Any]]:
    # Get all click command names in module 'click_module'
    click_command_names = [c[0] for c in inspect.getmembers(click_module, is_click_command)]
    # Get all function names in module 'api_module'
    api_function_names = [f[0] for f in inspect.getmembers(api_module, inspect.isfunction)
                          if f[1].__cli_function__]
    return click_command_names, api_function_names



import math
import time
from pathlib import Path
from threading import Thread

from docker.models.containers import Container

from exasol_integration_test_docker_environment.lib.base.still_running_logger import StillRunningLogger
from exasol_integration_test_docker_environment.lib.logging.container_log_handler import ContainerLogHandler


class DBContainerLogThread(Thread):
    def __init__(self, container: Container, logger, log_file: Path, description: str):
        super().__init__()
        self.complete_log = []
        self.description = description
        self.logger = logger
        self.log_file = log_file
        self.container = container
        self.finish = False
        self.previous_timestamp = None
        self.current_timestamp = None
        self.error_message = None

    def stop(self):
        self.logger.info("Stop ContainerLogThread")
        self.finish = True

    def run(self):
        with ContainerLogHandler(self.log_file, self.logger, self.description) as log_handler:
            still_running_logger = StillRunningLogger(
                self.logger, self.description)
            while not self.finish:
                self.current_timestamp = math.floor(time.time())
                log = self.container.logs(since=self.previous_timestamp, until=self.current_timestamp)
                if len(log) != 0:
                    still_running_logger.log()
                    log_handler.handle_log_lines(log)
                log_line = log.decode("utf-8").lower()
                if ("error" in log_line and not "sshd was not started" in log_line) \
                        or "exception" in log_line \
                        or ("returned with state 1" in log_line 
                                and not "(membership) returned with state 1" in log_line): # exclude webui not found in 7.0.0 
                    self.logger.info("ContainerLogHandler error message, %s", log_line)
                    self.error_message = log_line
                    self.finish = True
                self.previous_timestamp = self.current_timestamp
                self.complete_log = log_handler.get_complete_log().copy()
                time.sleep(1)



import luigi

from exasol_integration_test_docker_environment.lib.base.dependency_logger_base_task import DependencyLoggerBaseTask
from exasol_integration_test_docker_environment.lib.data.environment_type import EnvironmentType
from exasol_integration_test_docker_environment.lib.test_environment.parameter.spawn_test_environment_parameter import \
    SpawnTestEnvironmentParameter
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_environment_with_docker_db import \
    SpawnTestEnvironmentWithDockerDB
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_environment_with_external_db import \
    SpawnTestEnvironmentWithExternalDB


class SpawnTestEnvironment(DependencyLoggerBaseTask, SpawnTestEnvironmentParameter):
    DEFAULT_DB_USER = "sys"
    DEFAULT_DATABASE_PASSWORD = "exasol"
    DEFAULT_BUCKETFS_WRITE_PASSWORD = "write"
    environment_name = luigi.Parameter()

    def register_required(self):
        task = self._create_spawn_environment_task()
        self._environment_info_future = self.register_dependency(task)

    def _create_spawn_environment_task(self):
        if self.environment_type == EnvironmentType.docker_db:
            return self._create_docker_db_environment()
        else:
            return self._create_external_db_environment()

    def _create_external_db_environment(self):
        if self.external_exasol_db_host is None:
            raise Exception("external_exasol_db_host not set")
        if self.external_exasol_db_port is None:
            raise Exception("external_exasol_db_port not set")
        if self.external_exasol_bucketfs_port is None:
            raise Exception("external_exasol_bucketfs_port not set")
        task = \
            self.create_child_task_with_common_params(
                SpawnTestEnvironmentWithExternalDB,
                db_user=self.external_exasol_db_user,
                db_password=self.external_exasol_db_password,
                bucketfs_write_password=self.external_exasol_bucketfs_write_password
            )
        return task

    def _create_docker_db_environment(self):
        task = \
            self.create_child_task_with_common_params(
                SpawnTestEnvironmentWithDockerDB,
                db_user=self.DEFAULT_DB_USER,
                db_password=self.DEFAULT_DATABASE_PASSWORD,
                bucketfs_write_password=self.DEFAULT_BUCKETFS_WRITE_PASSWORD
            )
        return task

    def run_task(self):
        environment_info = self.get_values_from_future(self._environment_info_future)
        self.return_object(environment_info)



from exasol_integration_test_docker_environment.lib.base.info import Info
from exasol_integration_test_docker_environment.lib.data.container_info import ContainerInfo


class DatabaseInfo(Info):

    def __init__(self, host: str, db_port: str, bucketfs_port: str,
                 reused: bool, container_info: ContainerInfo = None):
        self.container_info = container_info
        self.bucketfs_port = bucketfs_port
        self.db_port = db_port
        self.host = host
        self.reused = reused



import luigi
from luigi import Config


class DatabaseCredentials:
    def __init__(self, db_user: str, db_password: str, bucketfs_write_password: str):
        self.bucketfs_write_password = bucketfs_write_password
        self.db_password = db_password
        self.db_user = db_user


class DatabaseCredentialsParameter(Config):
    db_user = luigi.Parameter()
    db_password = luigi.Parameter(significant=False,
                                  visibility=luigi.parameter.ParameterVisibility.HIDDEN)
    bucketfs_write_password = luigi.Parameter(significant=False,
                                              visibility=luigi.parameter.ParameterVisibility.HIDDEN)

    def get_database_credentials(self):
        return DatabaseCredentials(self.db_user, self.db_password, self.bucketfs_write_password)



from pathlib import Path
from typing import Any

import jsonpickle
from luigi import LocalTarget


class JsonPickleTarget(LocalTarget):

    def __init__(self, path: Path, is_tmp: bool = False):
        super().__init__(path=str(path), is_tmp=is_tmp)

    def write(self, obj:Any, indent:int=None):
        jsonpickle.set_preferred_backend('simplejson')
        jsonpickle.set_encoder_options('simplejson', indent=indent)
        json_str=jsonpickle.encode(obj)
        with self.open("w") as f:
            f.write(json_str)

    def read(self):
        jsonpickle.set_preferred_backend('simplejson')
        with self.open("r") as f:
            json_str = f.read()
            obj = jsonpickle.decode(json_str)
            return obj



import gzip
import pathlib
import time
from datetime import datetime, timedelta

import luigi
from docker.models.containers import Container

from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.data.database_credentials import DatabaseCredentialsParameter
from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.db_container_log_thread import \
    DBContainerLogThread
from exasol_integration_test_docker_environment.lib.test_environment.database_waiters.is_database_ready_thread import \
    IsDatabaseReadyThread


class WaitForTestDockerDatabase(DockerBaseTask, DatabaseCredentialsParameter):
    environment_name = luigi.Parameter()
    database_info = JsonPickleParameter(DatabaseInfo, significant=False)  # type: DatabaseInfo
    db_startup_timeout_in_seconds = luigi.IntParameter(10 * 60, significant=False)
    attempt = luigi.IntParameter(1)
    docker_db_image_version = luigi.Parameter()

    def run_task(self):
        with self._get_docker_client() as docker_client:
            db_container_name = self.database_info.container_info.container_name
            db_container = docker_client.containers.get(db_container_name)
            is_database_ready = \
                self.wait_for_database_startup(db_container)
            after_startup_db_log_file = self.get_log_path().joinpath("after_startup_db_log.tar.gz")
            self.save_db_log_files_as_gzip_tar(after_startup_db_log_file, db_container)
            self.return_object(is_database_ready)

    def wait_for_database_startup(self,
                                  db_container: Container):
        container_log_thread, is_database_ready_thread = \
            self.start_wait_threads(db_container)
        is_database_ready = \
            self.wait_for_threads(container_log_thread, is_database_ready_thread)
        self.join_threads(container_log_thread, is_database_ready_thread)
        return is_database_ready

    def start_wait_threads(self, db_container: Container):
        startup_log_file = self.get_log_path().joinpath("startup.log")
        container_log_thread = DBContainerLogThread(db_container,
                                                    self.logger,
                                                    startup_log_file,
                                                    "Database Startup %s" % db_container.name)
        container_log_thread.start()
        is_database_ready_thread = IsDatabaseReadyThread(self.logger,
                                                         self.database_info,
                                                         db_container,
                                                         self.get_database_credentials(),
                                                         self.docker_db_image_version)
        is_database_ready_thread.start()
        return container_log_thread, is_database_ready_thread

    def join_threads(self, container_log_thread: DBContainerLogThread,
                     is_database_ready_thread: IsDatabaseReadyThread):
        container_log_thread.stop()
        is_database_ready_thread.stop()
        container_log_thread.join()
        is_database_ready_thread.join()

    def wait_for_threads(self, container_log_thread: DBContainerLogThread,
                         is_database_ready_thread: IsDatabaseReadyThread):
        is_database_ready = False
        reason = None
        start_time = datetime.now()
        while (True):
            if container_log_thread.error_message != None:
                is_database_ready = False
                reason = "error message in container log"
                break
            if is_database_ready_thread.finish and is_database_ready_thread.is_ready:
                is_database_ready = True
                break
            elif is_database_ready_thread.finish:
                reason = "error in is_database_ready_thread"
                break
            if self.timeout_occured(start_time):
                reason = f"timeout after after {self.db_startup_timeout_in_seconds} seconds"
                is_database_ready = False
                break
            time.sleep(1)
        if not is_database_ready:
            self.log_database_not_ready(container_log_thread, is_database_ready_thread, reason)
        container_log_thread.stop()
        is_database_ready_thread.stop()
        return is_database_ready

    def log_database_not_ready(self, container_log_thread: DBContainerLogThread,
                               is_database_ready_thread: IsDatabaseReadyThread, reason):
        container_log = '\n'.join(container_log_thread.complete_log)
        log_information = f"""
========== IsDatabaseReadyThread output db connection: ============
{is_database_ready_thread.output_db_connection}
========== IsDatabaseReadyThread output bucketfs connection: ============
{is_database_ready_thread.output_bucketfs_connection}
========== Container-Log: ============ 
{container_log}
"""
        self.logger.warning(
            'Database startup failed for following reason "%s", here some debug information \n%s',
            reason, log_information)

    def timeout_occured(self, start_time):
        timeout = timedelta(seconds=self.db_startup_timeout_in_seconds)
        return datetime.now() - start_time > timeout

    def save_db_log_files_as_gzip_tar(self, path: pathlib.Path, database_container: Container):
        stream, stat = database_container.get_archive("/exa/logs")
        with gzip.open(path, "wb") as file:
            for chunk in stream:
                file.write(chunk)

    def write_output(self, is_database_ready: bool):
        with self.output().open("w") as file:
            file.write(str(is_database_ready))



from typing import Tuple, Optional

from exasol_integration_test_docker_environment.lib.api.common import set_docker_repository_config, \
    run_task, set_build_config, generate_root_task, no_cli_function
from exasol_integration_test_docker_environment.cli.options.docker_repository_options import \
    DEFAULT_DOCKER_REPOSITORY_NAME
from exasol_integration_test_docker_environment.cli.options.system_options import DEFAULT_OUTPUT_DIRECTORY
from exasol_integration_test_docker_environment.lib.data.test_container_content_description import \
    TestContainerContentDescription
from exasol_integration_test_docker_environment.lib.docker.images.image_info import ImageInfo
from exasol_integration_test_docker_environment.lib.test_environment.analyze_test_container import \
    AnalyzeTestContainer, DockerTestContainerPush


@no_cli_function
def push_test_container(
        test_container_content: TestContainerContentDescription,
        force_push: bool = False,
        push_all: bool = False,
        force_rebuild: bool = False,
        force_rebuild_from: Tuple[str, ...] = tuple(),
        force_pull: bool = False,
        output_directory: str = DEFAULT_OUTPUT_DIRECTORY,
        temporary_base_directory: str = "/tmp",
        log_build_context_content: bool = False,
        cache_directory: Optional[str] = None,
        build_name: Optional[str] = None,
        source_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
        source_docker_tag_prefix: str = '',
        source_docker_username: Optional[str] = None,
        source_docker_password: Optional[str] = None,
        target_docker_repository_name: str = DEFAULT_DOCKER_REPOSITORY_NAME,
        target_docker_tag_prefix: str = '',
        target_docker_username: Optional[str] = None,
        target_docker_password: Optional[str] = None,
        workers: int = 5,
        task_dependencies_dot_file: Optional[str] = None,
        log_level: Optional[str] = None,
        use_job_specific_log_file: bool = False
) -> ImageInfo:
    """
    This function pushes all stages of the test container for the test environment.
    If the stages do not exist locally, the system will build or pull them before the push.
    """
    set_build_config(force_rebuild,
                     force_rebuild_from,
                     force_pull,
                     log_build_context_content,
                     output_directory,
                     temporary_base_directory,
                     cache_directory,
                     build_name)
    # Use AnalyzeTestContainer to ensure that all luigi processes got it loaded
    analyze_task = AnalyzeTestContainer.__class__.__name__

    set_docker_repository_config(source_docker_password, source_docker_repository_name, source_docker_username,
                                 source_docker_tag_prefix, "source")
    set_docker_repository_config(target_docker_password, target_docker_repository_name, target_docker_username,
                                 target_docker_tag_prefix, "target")
    task_creator = lambda: generate_root_task(task_class=DockerTestContainerPush,
                                              test_container_content=test_container_content,
                                              force_push=force_push,
                                              push_all=push_all)
    image_infos = run_task(task_creator, workers, task_dependencies_dot_file,
                           log_level=log_level,
                           use_job_specific_log_file=use_job_specific_log_file)
    return image_infos[0]



from exasol_integration_test_docker_environment.lib.base.info import Info


class DockerNetworkInfo(Info):

    def __init__(self, network_name: str, subnet: str, gateway: str, reused: bool = False):
        self.gateway = gateway
        self.subnet = subnet
        self.network_name = network_name
        self.reused = reused



from typing import Optional

from exasol_integration_test_docker_environment.cli.options.test_environment_options import LATEST_DB_VERSION

DEFAULT_VERSION = "default"


class DbVersion:
    def __init__(self, major, minor, stable):
        self.major = major
        self.minor = minor
        self.stable = stable

    @classmethod
    def from_db_version_str(cls, db_version_str: Optional[str]):
        db_version = db_version_str
        if db_version_str in (None, DEFAULT_VERSION):
            db_version = LATEST_DB_VERSION
        if db_version.endswith("-d1"):
            db_version = "-".join(db_version.split("-")[0:-1])
        if db_version.startswith("prerelease-"):
            db_version = "-".join(db_version.split("-")[1:])
        version = tuple([int(v) for v in db_version.split(".")])
        if len(version) != 3:
            raise ValueError(f"Invalid db version given: {db_version}")
        return DbVersion(version[0], version[1], version[2])

    def __str__(self):
        return f"{self.major}.{self.minor}.{self.stable}"

    def to_tuple(self):
        return self.major, self.minor, self.stable

    def __eq__(self, other) -> bool:
        return self.to_tuple() == other.to_tuple()

    def __lt__(self, other) -> bool:
        return self.to_tuple() < other.to_tuple()

    def __le__(self, other) -> bool:
        return self.to_tuple() <= other.to_tuple()

    def __ge__(self, other) -> bool:
        return self.to_tuple() >= other.to_tuple()

    def __gt__(self, other) -> bool:
        return self.to_tuple() > other.to_tuple()


def db_version_supports_custom_certificates(db_version_str: Optional[str]) -> bool:
    return DbVersion.from_db_version_str(db_version_str) > DbVersion(7, 0, 5)



from luigi.parameter import ParameterVisibility

from exasol_integration_test_docker_environment.lib.base.json_pickle_parameter import JsonPickleParameter
from exasol_integration_test_docker_environment.lib.data.test_container_content_description import \
    TestContainerContentDescription


class TestContainerParameter:
    test_container_content = JsonPickleParameter(TestContainerContentDescription,
                                                 visibility=ParameterVisibility.HIDDEN)


class OptionalTestContainerParameter:
    test_container_content = JsonPickleParameter(TestContainerContentDescription,
                                                 visibility=ParameterVisibility.HIDDEN,
                                                 is_optional=True)



from enum import Enum
from sys import stderr

import jsonpickle


class TaskDescription:

    def __init__(self, id: str, representation: str):
        self.representation = representation
        self.id = id

    @property
    def formatted_representation(self):
        representation = self.representation.replace('"', '\\"')
        return f'"{representation}"'

    @property
    def formatted_id(self):
        assert '"' not in self.id
        return f'"{self.id}"'


class DependencyType(Enum):
    requires = 1,
    dynamic = 2


class DependencyState(Enum):
    requested = 1
    finished = 2


class TaskDependency:

    def __init__(self, source: TaskDescription, target: TaskDescription,
                 type: DependencyType, index: int, state: DependencyState):
        self.state = state.name
        self.index = index
        self.type = type.name
        self.target = target
        self.source = source

    def to_json(self):
        jsonpickle.set_preferred_backend('simplejson')
        jsonpickle.set_encoder_options('simplejson', sort_keys=True)
        return jsonpickle.encode(self)

    @classmethod
    def from_json(cls, json_string: object) -> object:
        loaded_object = jsonpickle.decode(json_string)
        if not isinstance(loaded_object, cls):
            raise TypeError("Type %s of loaded object does not match %s" % (type(loaded_object), cls))
        return loaded_object

    @property
    def formatted(self):
        assert '"' not in str(self)
        return f'"{str(self)}"'

    def __str__(self):
        return (
                    f"TaskDependency(source={self.source}, "
                    f"target={self.target}, type={self.type}, " 
                    f"index={self.index}, state={self.state})"
               )




import itertools
from typing import Dict, List

from exasol_integration_test_docker_environment.abstract_method_exception import AbstractMethodException
from exasol_integration_test_docker_environment.lib.base.base_task import BaseTask
from exasol_integration_test_docker_environment.lib.docker.images.create.docker_image_create_task import \
    DockerCreateImageTask, DockerCreateImageTaskWithDeps
from exasol_integration_test_docker_environment.lib.docker.images.required_task_info import RequiredTaskInfo


class TaskCreatorFromBuildTasks:

    def create_tasks_for_build_tasks(self, build_tasks: Dict[str, DockerCreateImageTask]) \
            -> List[BaseTask]:
        tasks_per_goal = [self._create_tasks_for_build_task(build_task)
                          for goal, build_task in build_tasks.items()]
        return list(itertools.chain.from_iterable(tasks_per_goal))

    def _create_tasks_for_build_task(self, build_task: DockerCreateImageTask) \
            -> List[BaseTask]:
        if isinstance(build_task, DockerCreateImageTaskWithDeps):
            tasks = self.create_tasks_for_build_tasks(build_task.required_tasks)
            task = self._create_task(build_task)
            return [task] + tasks
        else:
            task = self._create_task(build_task)
            return [task]

    def _create_task(self, build_task):
        required_task_info = self._create_required_task_info(build_task)
        task = self.create_task_with_required_tasks(build_task, required_task_info)
        return task

    def _create_required_task_info(self, build_task: DockerCreateImageTask):
        required_task_info = \
            RequiredTaskInfo(module_name=build_task.__module__,
                             class_name=build_task.__class__.__name__,
                             params=build_task.param_kwargs)
        return required_task_info

    def create_task_with_required_tasks(self, build_task, required_task_info) -> BaseTask:
        raise AbstractMethodException()



from pathlib import Path
from typing import Generator, Tuple, Optional

import luigi

from exasol_integration_test_docker_environment.abstract_method_exception import AbstractMethodException
from exasol_integration_test_docker_environment.lib.base.base_task import BaseTask
from exasol_integration_test_docker_environment.lib.base.docker_base_task import DockerBaseTask
from exasol_integration_test_docker_environment.lib.data.container_info import ContainerInfo
from exasol_integration_test_docker_environment.lib.data.database_credentials import DatabaseCredentialsParameter
from exasol_integration_test_docker_environment.lib.data.database_info import DatabaseInfo
from exasol_integration_test_docker_environment.lib.data.docker_network_info import DockerNetworkInfo
from exasol_integration_test_docker_environment.lib.data.docker_volume_info import DockerVolumeInfo
from exasol_integration_test_docker_environment.lib.data.environment_info import EnvironmentInfo
from exasol_integration_test_docker_environment.lib.test_environment.docker_container_copy import DockerContainerCopy
from exasol_integration_test_docker_environment.lib.test_environment.parameter.general_spawn_test_environment_parameter import \
    GeneralSpawnTestEnvironmentParameter
from exasol_integration_test_docker_environment.lib.test_environment.spawn_test_container import SpawnTestContainer

DATABASE = "database"

TEST_CONTAINER = "test_container"


class AbstractSpawnTestEnvironment(DockerBaseTask,
                                   GeneralSpawnTestEnvironmentParameter,
                                   DatabaseCredentialsParameter):
    environment_name = luigi.Parameter()  # type: str

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.test_container_name = f"""test_container_{self.environment_name}"""
        self.network_name = f"""db_network_{self.environment_name}"""

    def get_environment_type(self):
        raise AbstractMethodException()

    def run_task(self):
        test_environment_info = yield from self._attempt_database_start()
        self.return_object(test_environment_info)

    def _attempt_database_start(self):
        is_database_ready = False
        attempt = 0
        database_info = None
        test_container_info = None
        while not is_database_ready and attempt < self.max_start_attempts:
            network_info, database_info, is_database_ready, test_container_info = \
                yield from self._start_database(attempt)
            attempt += 1
        if not is_database_ready and not attempt < self.max_start_attempts:
            raise Exception(f"Maximum attempts {attempt} to start the database reached.")
        test_environment_info = \
            EnvironmentInfo(name=self.environment_name,
                            env_type=self.get_environment_type(),
                            database_info=database_info,
                            test_container_info=test_container_info,
                            network_info=network_info)
        self.create_test_environment_info_in_test_container_and_on_host(test_environment_info)
        return test_environment_info

    def create_test_environment_info_in_test_container(self, test_environment_info: EnvironmentInfo,
                                                       environment_variables: str,
                                                       environment_variables_with_export: str,
                                                       json: str):
        test_container_name = test_environment_info.test_container_info.container_name
        with self._get_docker_client() as docker_client:
            test_container = docker_client.containers.get(test_container_name)
            self.logger.info(f"Create test environment info in test container '{test_container_name}' at '/'")
            copy = DockerContainerCopy(test_container)
            copy.add_string_to_file("environment_info.json", json)
            copy.add_string_to_file("environment_info.conf", environment_variables)
            copy.add_string_to_file("environment_info.sh", environment_variables_with_export)
            copy.copy("/")

    def create_test_environment_info_in_test_container_and_on_host(
            self, test_environment_info: EnvironmentInfo):
        test_environment_info_base_host_path = Path(self.get_cache_path(),
                                                    f"environments/{self.environment_name}")
        test_environment_info_base_host_path.mkdir(exist_ok=True, parents=True)
        self.logger.info(f"Create test environment info on the host at '{test_environment_info_base_host_path}'")

        json = test_environment_info.to_json()
        cache_environment_info_json_path = Path(test_environment_info_base_host_path,
                                                "environment_info.json")
        with cache_environment_info_json_path.open("w") as f:
            f.write(json)

        if test_environment_info.test_container_info is not None:
            test_container_name = test_environment_info.test_container_info.container_name
        else:
            test_container_name = ""
        environment_variables = \
            self.collect_environment_info_variables(test_container_name,
                                                    test_environment_info)
        cache_environment_info_conf_path = Path(test_environment_info_base_host_path,
                                                "environment_info.conf")
        with cache_environment_info_conf_path.open("w") as f:
            f.write(environment_variables)

        environment_variables_with_export = ""
        for line in environment_variables.splitlines():
            environment_variables_with_export += f"export {line}\n"
        cache_environment_info_sh_path = Path(test_environment_info_base_host_path, "environment_info.sh")
        with cache_environment_info_sh_path.open("w") as f:
            f.write(environment_variables_with_export)

        if test_environment_info.test_container_info is not None:
            self.create_test_environment_info_in_test_container(test_environment_info,
                                                                environment_variables,
                                                                environment_variables_with_export, json)

    def collect_environment_info_variables(self, test_container_name: str, test_environment_info):
        environment_variables = ""
        environment_variables += f"ENVIRONMENT_NAME={test_environment_info.name}\n"
        environment_variables += f"ENVIRONMENT_TYPE={test_environment_info.type}\n"
        environment_variables += f"ENVIRONMENT_DATABASE_HOST={test_environment_info.database_info.host}\n"
        environment_variables += f"ENVIRONMENT_DATABASE_DB_PORT={test_environment_info.database_info.db_port}\n"
        environment_variables += f"ENVIRONMENT_DATABASE_BUCKETFS_PORT={test_environment_info.database_info.bucketfs_port}\n"
        if test_environment_info.database_info.container_info is not None:
            environment_variables += f"""ENVIRONMENT_DATABASE_CONTAINER_NAME={test_environment_info.database_info.container_info.container_name}\n"""
            database_container_network_aliases = " ".join(
                test_environment_info.database_info.container_info.network_aliases)
            environment_variables += f"""ENVIRONMENT_DATABASE_CONTAINER_NETWORK_ALIASES="{database_container_network_aliases}"\n"""
            environment_variables += f"""ENVIRONMENT_DATABASE_CONTAINER_IP_ADDRESS={test_environment_info.database_info.container_info.ip_address}\n"""
            environment_variables += f"""ENVIRONMENT_DATABASE_CONTAINER_VOLUMNE_NAME={test_environment_info.database_info.container_info.volume_name}\n"""
            with self._get_docker_client() as docker_client:
                db_container = docker_client.containers.get(
                    test_environment_info.database_info.container_info.container_name)
                db_container.reload()
                default_bridge_ip_address = db_container.attrs["NetworkSettings"]["Networks"]["bridge"]["IPAddress"]
            environment_variables += f"""ENVIRONMENT_DATABASE_CONTAINER_DEFAULT_BRIDGE_IP_ADDRESS={default_bridge_ip_address}\n"""
        if test_environment_info.test_container_info is not None:
            environment_variables += f"""ENVIRONMENT_TEST_CONTAINER_NAME={test_container_name}\n"""
            test_container_network_aliases = " ".join(test_environment_info.test_container_info.network_aliases)
            environment_variables += f"""ENVIRONMENT_TEST_CONTAINER_NETWORK_ALIASES="{test_container_network_aliases}"\n"""
            environment_variables += f"""ENVIRONMENT_TEST_CONTAINER_IP_ADDRESS={test_environment_info.test_container_info.ip_address}\n"""
        return environment_variables

    def _start_database(self, attempt) \
            -> Generator[BaseTask, BaseTask, Tuple[DockerNetworkInfo, DatabaseInfo, bool, Optional[ContainerInfo]]]:
        network_info = yield from self._create_network(attempt)
        ssl_volume_info = None
        if self.create_certificates:
            ssl_volume_info = yield from self._create_ssl_certificates()
        database_info, test_container_info = \
            yield from self._spawn_database_and_test_container(network_info, ssl_volume_info, attempt)
        is_database_ready = yield from self._wait_for_database(database_info, attempt)
        return network_info, database_info, is_database_ready, test_container_info

    def _create_ssl_certificates(self) -> DockerVolumeInfo:
        ssl_info_future = yield from self.run_dependencies(self.create_ssl_certificates())
        ssl_info = self.get_values_from_future(ssl_info_future)
        return ssl_info

    def create_ssl_certificates(self):
        raise AbstractMethodException()

    def _create_network(self, attempt):
        network_info_future = yield from self.run_dependencies(self.create_network_task(attempt))
        network_info = self.get_values_from_future(network_info_future)
        return network_info

    def create_network_task(self, attempt: int):
        raise AbstractMethodException()

    def _spawn_database_and_test_container(self,
                                           network_info: DockerNetworkInfo,
                                           certificate_volume_info: Optional[DockerVolumeInfo],
                                           attempt: int) -> Tuple[DatabaseInfo, Optional[ContainerInfo]]:
        certificate_volume_name = certificate_volume_info.volume_name if certificate_volume_info is not None else None
        dependencies_tasks = {
                DATABASE: self.create_spawn_database_task(network_info, certificate_volume_info, attempt)
            }
        if self.test_container_content is not None:
            dependencies_tasks[TEST_CONTAINER] = \
                self.create_spawn_test_container_task(network_info, certificate_volume_name, attempt)
        database_and_test_container_info_future = yield from self.run_dependencies(dependencies_tasks)
        database_and_test_container_info = \
            self.get_values_from_futures(database_and_test_container_info_future)
        test_container_info = None
        if self.test_container_content is not None:
            test_container_info = database_and_test_container_info[TEST_CONTAINER]
        database_info = database_and_test_container_info[DATABASE]
        return database_info, test_container_info

    def create_spawn_database_task(self,
                                   network_info: DockerNetworkInfo,
                                   certificate_volume_info: Optional[DockerVolumeInfo],
                                   attempt: int):
        raise AbstractMethodException()

    def create_spawn_test_container_task(self, network_info: DockerNetworkInfo,
                                         certificate_volume_name: str, attempt: int):
        return self.create_child_task_with_common_params(
                SpawnTestContainer,
                test_container_name=self.test_container_name,
                network_info=network_info,
                ip_address_index_in_subnet=1,
                certificate_volume_name=certificate_volume_name,
                attempt=attempt,
                test_container_content=self.test_container_content
                )

    def _wait_for_database(self,
                           database_info: DatabaseInfo,
                           attempt: int):
        database_ready_target_future = \
            yield from self.run_dependencies(self.create_wait_for_database_task(attempt, database_info))
        is_database_ready = self.get_values_from_futures(database_ready_target_future)
        return is_database_ready

    def create_wait_for_database_task(self,
                                      attempt: int,
                                      database_info: DatabaseInfo):
        raise AbstractMethodException()



import click

DEFAULT_DOCKER_REPOSITORY_NAME = 'exasol/script-language-container'

docker_repository_options = [
    click.option('--source-docker-repository-name', type=str,
                 default=DEFAULT_DOCKER_REPOSITORY_NAME,
                 show_default=True,
                 help="Name of the docker repository for pulling cached stages. "
                      "The repository name may contain the URL of the docker registry, "
                      "the username and the actual repository name. "
                      "A common structure is <docker-registry-url>/<username>/<repository-name>"),
    click.option('--source-docker-tag-prefix', type=str,
                 default="",
                 show_default=True,
                 help="Prefix for the tags which are used for pulling of cached stages"),
    click.option('--source-docker-username', type=str,
                 help="Username for the docker registry from where the system pulls cached stages.",
                 required=False),
    click.option('--source-docker-password', type=str,
                 help="Password for the docker registry from where the system pulls cached stages. "
                      "Without password option the system prompts for the password."),
    click.option('--target-docker-repository-name', type=str,
                 default=DEFAULT_DOCKER_REPOSITORY_NAME,
                 show_default=True,
                 help="Name of the docker repository for naming and pushing images of stages. "
                      "The repository name may contain the URL of the docker registry, "
                      "the username and the actual repository name. "
                      "A common structure is <docker-registry-url>/<username>/<repository-name>"),
    click.option('--target-docker-tag-prefix', type=str,
                 default="",
                 show_default=True,
                 help="Prefix for the tags which are used for naming and pushing of stages"),
    click.option('--target-docker-username', type=str,
                 help="Username for the docker registry where the system pushes images of stages.",
                 required=False),
    click.option('--target-docker-password', type=str,
                 help="Password for the docker registry where the system pushes images of stages. "
                      "Without password option the system prompts for the password."),
]

simple_docker_repository_options = [
    click.option('--docker-repository-name', type=str,
                 default=DEFAULT_DOCKER_REPOSITORY_NAME,
                 show_default=True,
                 help="Name of the docker repository for naming images. "
                      "The repository name may contain the URL of the docker registry, "
                      "the username and the actual repository name. "
                      "A common structure is <docker-registry-url>/<username>/<repository-name>"),
    click.option('--docker-tag-prefix', type=str,
                 default="",
                 show_default=True,
                 help="Prefix for the tags of the images"),
]



import luigi
from luigi import Config
from luigi.parameter import ParameterVisibility


class DockerParameter(Config):
    """
    Docker parameters used for Tasks accessing Docker client.
    """

    timeout = luigi.IntParameter(100000, significant=False, visibility=ParameterVisibility.PRIVATE)
    no_cache = luigi.BoolParameter(False)



import pickle
from pathlib import Path

from luigi import LocalTarget
from luigi.format import NopFormat

FORMAT = NopFormat()

class PickleTarget(LocalTarget):

    def __init__(self, path: Path, is_tmp: bool = False):
        super().__init__(path=str(path), is_tmp=is_tmp, format=FORMAT)

    def write(self, obj):
        with self.open("w") as f:
            pickle.dump(obj, f)

    def read(self):
        with self.open("r") as f:
            obj = pickle.load(f)
            return obj



import json
import logging
import time

import requests

from exasol_integration_test_docker_environment.lib.docker import ContextDockerClient
from exasol_integration_test_docker_environment.lib.docker.container.utils import remove_docker_container
from exasol_integration_test_docker_environment.testing.utils import find_free_ports


def default_docker_repository_name(env_name: str) -> str:
    return f"exaslct_test/{env_name.lower()}"


class LocalDockerRegistry:
    def __init__(self, name: str, registry_container, registry_port):
        self._name = name
        self._registry_container = registry_container
        self._registry_port = registry_port

    @property
    def name(self):
        return f"localhost:{self._registry_port}/{self._name.lower()}"

    @property
    def images(self):
        url = f"http://localhost:{self._registry_port}/v2/{self._name.lower()}/tags/list"
        result = requests.request("GET", url)
        images = json.loads(result.content.decode("UTF-8"))
        return images

    @property
    def repositories(self):
        result = requests.request("GET", f"http://localhost:{self._registry_port}/v2/_catalog/")
        repositories_ = json.loads(result.content.decode("UTF-8"))["repositories"]
        return repositories_


class LocalDockerRegistryContextManager:
    def __init__(self, name: str):
        self._name = name
        self._local_docker_registry = None

    def __enter__(self):
        registry_port = find_free_ports(1)[0]
        registry_container_name = self._name.replace("/", "_") + "_registry"
        with ContextDockerClient() as docker_client:
            logging.debug("Start pull of registry:2")
            docker_client.images.pull(repository="registry", tag="2")
            logging.debug(f"Start container of {registry_container_name}")
            try:
                docker_client.containers.get(registry_container_name).remove(force=True)
            except:
                pass
            registry_container = docker_client.containers.run(
                image="registry:2", name=registry_container_name,
                ports={5000: registry_port},
                detach=True
            )
            time.sleep(10)
            logging.debug(f"Finished start container of {registry_container_name}")
            self._local_docker_registry = LocalDockerRegistry(self._name, registry_container, registry_port)
            return self._local_docker_registry

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.close()

    def close(self):
        if self._local_docker_registry is not None:
            remove_docker_container([self._local_docker_registry._registry_container.id])
